# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from .mpl import (
    plot_bode,
    plot_circuit,
    plot_complex_impedance,
    plot_data,
    plot_drt,
    plot_exploratory_tests,
    plot_fit,
    plot_gamma,
    plot_imaginary_impedance,
    plot_impedance_magnitude,
    plot_impedance_phase,
    plot_mu_xps,
    plot_nyquist,
    plot_real_impedance,
    plot_residual,
)
